// Jason Rohrer
// TransferThread.java

/**
*
* 	Abstract superclass for threads that transfer files,
* 	display status about transfers in gui, and get user input from GUI.
*
*	Created 5-11-2000
*	Mods:
*		Jason Rohrer	6-14-2000	Changed the way setProgress works.  Fixed 
*									a bug, and also improved readability of 
*									progressbar display.
*		Jason Rohrer	6-18-2000	Fixed setProgress so that KB/sec is displayed
*									correctly for resumed transfers.
*		Jason Rohrer	7-3-2000	Changed to use UnitsPrefixFormat to format file size and progress bar string.
*									Fixed so that DL and UL subclasses calculate the same time for a transfer
*									DL thread used to include time taken by initiating connection, since
*									start time was calculated when thread was run, not when transfer actually
*									started.
*		Jason Rohrer	7-23-2000	Changed so that all TransferThreads are set to 
*									minimum priority to make sure that the GUI
*									is always completely responsive.  Note that this
*									priority setting doesn't usually increase GUI responsiveness
*									noticibly, since TransferThreads are handling network connections
*									and are therefore blocked often enough so that the GUI thread
*									has ample time to run.  However, on faster network connections,
*									e.g., gigabit ethernet, the network threads will not
*									be blocked as often and could cause GUI responsiveness
*									to be sporadic.  We're assuming that as long as the GUI
*									isn't being manipulated or updated, MIN_PRIORITY thread
*									performance will not be affected.  This should be verified.
*/

package konspire.client;

import konspire.common.FileDescriptor;

import konspire.common.UnitsPrefixFormat;
import konspire.common.BaseTwoUnitsPrefixFormat;
import java.text.NumberFormat;

import javax.swing.JButton;
import javax.swing.JProgressBar;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

/**
 * Abstract superclass for threads that transfer files,
 * display status about transfers in gui, and get user input from GUI.
 *
 * @author Jason Rohrer
 */
public abstract class TransferThread extends Thread {

	/**
	 * Descriptor for file that is being transfered by this thread.
	 */ 
	protected FileDescriptor f;
	
	// where status is displayed and user input gathered
	private TransferTable table;
	
	private String fileSizeString;
	private String baseUnitsString = "B";

	private JProgressBar progBar;	// progress of transfer
	
	private JButton cancelButton;	// button to cancel transfer
	
	private TableRow row;	// pointer to row in table where this thread is represented
	
	private ButtonActionListener buttonListen;
	
	private long startTime = 0;
	private long lastTime = 0;	// for calculating transfer rate
	private long startB = 0;	// starting point of transfer, for resumed transfers
	private boolean startBSet = false;		// startKB set on first call of setProgress
	
	
	private boolean failed = false;
	
	
	// note:  progressbar range must be set as ints (not longs) which is a pain for
	// tracking file transfer percentages (since ints only support 4 GiB files )
	private int progressBarSize = 1000;		// number of increments in progressbar
					
	
	/**
	 * Constructs a <code>TransferThread</code>.  Protected
	 * becase <code>TransferThread</code> is abstract.
	 *
	 * @param fDes descriptor of the file to be transfered by this thread
	 */
	protected TransferThread( FileDescriptor fDes ) {
		setPriority( Thread.MIN_PRIORITY );
		
		f = fDes;
		
		UnitsPrefixFormat prefixFormat = new BaseTwoUnitsPrefixFormat();
		
		fileSizeString = prefixFormat.convertUnitsPrefix( f.getSize(), 
									baseUnitsString );
		
		
		cancelButton = new JButton( "cancel" );
		
		progBar = new JProgressBar( 0, progressBarSize );
		progBar.setString( "connecting" );
		progBar.setStringPainted( true );
		
		buttonListen = new ButtonActionListener();
		
		cancelButton.addActionListener( buttonListen );
		}
	
	
	/**
	 * Sets the table where this thread will display itself
	 *
	 * @param t table where this thread will display itself
	 */
	public void setTable( TransferTable t ) {
		table = t;
		}
	
	
	/**
	 * Starts this thread.
	 */
	public void run() {
		// create and add row of objects representing this thread
		Object [] oArray = { f.getName(), fileSizeString, progBar, cancelButton };
		row = table.addRow( oArray );  
		
		
		
		startTransfer();
		
		setProgressDoneTime(); 
		
		transferDone();
		// don't remove row here
		}
	
	/**
	 * Sets final progress status message after transfer completes.
	 * Displays time taken by transfer.
	 */
	private void setProgressDoneTime() {
		
		long newTime = System.currentTimeMillis();
		long deltaTime = newTime - startTime;
		
		long secondsTime = deltaTime / 1000;
		
		if( !startBSet ) {
			// if startTime not defined, transfer failed during init
			secondsTime = 0;
			} 
		
		if( !failed ) {
			progBar.setString( "complete: " + secondsTime + " sec" );
			}
		else {
			progBar.setString( "failed: " + secondsTime + " sec" );
			}
		table.repaint();
		}
	
	/**
	 * Changes progress bar to reflect bytes transfered so far
	 *
	 * @param downloadedB bytes transfered so far
	 */
	protected void setProgress( long downloadedB ) {
		if( !startBSet ) {
			startB = downloadedB;
			startBSet = true;
			
			lastTime = System.currentTimeMillis();
			startTime = lastTime;
			
			// don't set progress value in first call, since lastTime = startTime
			return;
			}
		
		long newTime = System.currentTimeMillis();
		
		double BPerSec = (double)( downloadedB - startB ) / ( ( newTime - startTime ) * 0.001f );
		
		UnitsPrefixFormat prefixFormat = new BaseTwoUnitsPrefixFormat();
		
		String BPerSecString = prefixFormat.convertUnitsPrefix( BPerSec, 
									baseUnitsString );
		

		progBar.setString( BPerSecString );
		
		progBar.setValue( (int)( ( progressBarSize * downloadedB ) / f.getSize() ) );
		
		// Progress bar not updating unless frame resized
		// Attempted fix, 6-13-2000
		// Seems to work, 6-18-2000
		table.repaint();
		
		}
	
	
	/**
	 * Starts the file transfer.
	 * Returns when file transfer has stopped (completed or aborted)
	 */
	protected abstract void startTransfer();
	

	/**
	 * Ends the file transfer prematurely.
	 */
	protected abstract void abortTransfer();
	
	
	private boolean transferDone = false;
	
	/**
	 * Switches "cancel" button label to "clear" after transfer done
	 */
	protected void transferDone() {
		cancelButton.setText( "clear" );
		transferDone = true;
		table.repaint();
	}
	
	/**
	 * Called if the transfer fails in the run method.
	 */
	protected void transferFailed() {
		failed = true;
	}
	
	/**
	 * Listener for button events.
	 */
	private class ButtonActionListener implements ActionListener {
		public void actionPerformed( ActionEvent e ) {
			if( e.getSource() == cancelButton ) {
				
				if( !transferDone ) {
					// user chose to cancel this file transfer
					abortTransfer();
					// remove from table on cancel
					table.removeRow( row );
					}
				
				if( transferDone ) {
					// this button now used to remove thread from table
					table.removeRow( row );
					}
				
				}
			}
		}

	}