// Jason Rohrer
// Server.java

/**
*
*	Main server class for konspire
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	6-14-2000	Added support for passing client host name into 
*									database for file search. (used to filter results)
*		Jason Rohrer	7-4-2000	Fixed bug involving clients/servers that don't retrieve
*									their own IP Addresses correctly.  Previously, their FileHost
*									and NewServerUp messages did not contain a correct address 
*									(the localhost 127.0.0.1 address has even been witnessed).
*									Now, the server checks the included address against the address
*									attached to the other end of the socket.  The correct address is
*									passed back to the client/server in the ServerList message.
*/

package konspire.server;

import konspire.server.database.ClientDatabase;
import konspire.common.*;

import java.net.*;
import java.io.*;

import java.util.Date;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import konspire.common.Vector;

/**
 * Main server class for konspire.
 *
 * @author Jason Rohrer
 */
public class Server extends Thread {
	
	// database of all connected clients and files
	// network-wide, not just for clients connected to this server
	ClientDatabase database;
	
	// used to propagate messages to all known servers
	ServerPropagator propagator;
	
	// handles writing/reading server list on disk
	ServerListFile sListFile;
	
	Host localAddress;

	int maxClients;	// max clients to allow
	int maxConnections;	// max connections to backlog total
	int numClientsConnected = 0;
	
	static String hostsFileName = "servers.dat";

	/**
	 * Gets the network address of this <code>Server</code>.
	 *
	 * @return the network address of this <code>Server</code>
	 */
	public Host getLocalAddress() {
		return localAddress;
		}
	
	/**
	 * Constructs a <code>Server</code>.
	 *
	 * @param me local host 
	 * @param maxClie maximum number of simultaneous client connections
	 * @param newChain true iff this server is first in group of servers
	 *  (if true, this server won't try to connect to others to get the live server list)
	 */
	public Server( Host me, int maxClie, boolean newChain ) {
		
		database = new ClientDatabase();
		
		localAddress = me;
		
		maxClients = maxClie;
		maxConnections = 2 * maxClients;
		
		sListFile = new ServerListFile( hostsFileName );
		
		if( ! newChain ) {
			System.out.println( "Server will connect to an existing server chain." );
			System.out.println( "Reading server seed list from file: " + hostsFileName );
			Host[] suspectedOthers = sListFile.readHostsFromFile();
			
			
			
			boolean haveHostList = false;
			
			while( ! haveHostList ) {	// keep trying to get seed host from user until valid connection made
				if( suspectedOthers == null ) {
					System.out.println();
					System.out.println( "Please enter information about a known server." );
					suspectedOthers = getServerFromUser();
					}
				// get a list of actual servers from one of the seed servers
				System.out.println( "Trying to get a list of live servers." );
				Host[] knownOthers = lookForLiveServer( suspectedOthers );
				
				if( knownOthers != null ) {
					propagator = new ServerPropagator( knownOthers, sListFile, localAddress, this );
					System.out.println("Writing live server list to file (" + hostsFileName + ") for future use.");
					sListFile.writeHostsToFile( knownOthers );
					haveHostList = true;
					}
				else {
					System.out.println("No servers from provided host list found! (We're alone in this cold world.)");
					suspectedOthers = null;
					haveHostList = false;
					}
				}
			}
		else {		// new chain, don't look for server to receive server list from
			System.out.println( "Server will be first node in a new server chain." );
			propagator = new ServerPropagator( sListFile, localAddress, this );
			}
			
		}
	

	// overrides the Thread run() method

	public void run() {
		ServerSocket servSock;
		try {
			servSock = new ServerSocket( localAddress.getPort(), maxConnections );
			}
		catch( IOException e ) {
			System.out.println("Server socket failed to initialize.");
			return;
			}
		while( true ) {
			try {
				System.out.println( getTimeStamp() + ": Listening for connections." );
				Socket sock = servSock.accept();
				System.out.println( getTimeStamp() + ": Connection received." );
				
				// connection received
				// pass it to receiver thread
				ReceiveThread rThread = new ReceiveThread( sock );
				
				rThread.start();
				}
			catch( IOException e ) {
				System.out.println("Connection broken on server accept.");
				}
			}
		}	// end run()
		

		
	/**
	 * Processes command line arguments and constructs a <code>Server</code> at launch.
	 */
	public static void main (String[] args) {
		// retrieve user arguments and start the server thread

		Host serverLocalHost;
		
		String sAddress = "";
		int sPort = 5443;
		int sMaxConnect = 100;
		
		boolean newChain = false;	// should this server start up alone (w/o contacting others) ?
		
		//System.out.println("Number args in = " +args.length);
		
		if( args.length == 0 ) { 	// use defaults
			//System.out.print("Using default parameters, port: "+ s.port +", MaxConnect: "+ s.MaxConnect + "\n");
			
			// no defaults
			usage();
			}
		else if( args.length < 2 ) usage();
		else {
			// number of args <= 2
			try{
				sPort = Integer.parseInt(args[0]);
				sMaxConnect = Integer.parseInt(args[1]);
				if( args.length >= 3 ) {
					int newChainInt = Integer.parseInt( args[2] );
					if( newChainInt == 1 ) {
						newChain = true;
						}
					}
				}
			catch( NumberFormatException e ) {
				usage();
				}
			
			try {
				InetAddress iAddress = InetAddress.getLocalHost();
				sAddress = iAddress.getHostAddress();
				}
			catch( UnknownHostException e ) {
				System.out.println( "Failed to retrieve local host IP address." );
				System.exit(1);
				}
				
			System.out.print("Server at " + sAddress + ":" + sPort + ", MaxConnect: "+ sMaxConnect + "\n");
			}
		
		serverLocalHost = new Host( sAddress, sPort );
		
		System.out.println( "Initializing server." );
		Server s = new Server( serverLocalHost, sMaxConnect, newChain );
		
		
		System.out.println( "Starting server." );	
		s.start();
		
		System.out.println( "Enter q to shutdown server." );
		System.out.flush();
		
		try {
			while( System.in.read() != 'q' ) {
				// wait for user quit signal	
				}
			}
		catch( IOException e ) {
			System.out.println( "Exception occurred while reading user keyboard input: " + e.toString() );
			e.printStackTrace();
			}
		
		s.shutdown();
		}	// end main()
	
	
	/**
	 * Gets a <code>String</code> representation of the current date and time.
	 *
	 * @return <code>String</code> representing current date and time
	 */
	public String getTimeStamp() {
		Date current = new Date();
		return current.toString();
		}
	
	
	
	/**
	 * Kills the server.
	 */
	public void shutdown() {
		System.out.println( "Shutting down server." );
		
		// propagate a going down message
		
		ServerGoingDown m = new ServerGoingDown( localAddress );
		
		System.out.println( "Sending notification to server chain." );
		propagator.propagateMessage( m );
		
		System.out.println( "The server is halted." );
		try {
			sleep( 2000 );
			}
		catch( InterruptedException ie ) {
			}
		// exiting here seems to cause problems because some threads are still running...
		//System.exit(0);
		}
	
	/**
	 * Prompts the user to enter a seed server.
	 *
	 * @return an array of <code>Host</code> entered by the user, or <code>null</code>
	 * if retrieving input from user fails
	 */
	protected Host[] getServerFromUser() {
		try {
			BufferedReader userInput = new BufferedReader( new InputStreamReader( System.in ) );
			
			boolean hostNameRead = false;
			String hostName = "";
			
			while( ! hostNameRead ) {
				System.out.print( "Enter host name (or IP address): " );
				System.out.flush();
				
				String tempLine = null;
				
				tempLine = userInput.readLine();
				if( tempLine == null ) {
					System.out.println( "Keyboard input isn't working." );
					return null;
					}

				try {
					StringTokenizer hostTokenizer = new StringTokenizer( tempLine );
					hostName = hostTokenizer.nextToken();
					
					hostNameRead = true;
					}
				catch(  NoSuchElementException nsee ) {
					System.out.println("Invalid input string: " + tempLine );
					}
				}			

			int port = 0;
			boolean portRead = false;
			while( !portRead ) {
				System.out.print("Enter port number: " );
				System.out.flush();
				
				String tempLine = userInput.readLine();
				if( tempLine == null ) {
					System.out.println( "Keyboard input isn't working." );
					return null;
					}

				try{
					port = Integer.valueOf( tempLine ).intValue();
					portRead = true;
					}
				catch( NumberFormatException e ) {
					System.out.println("Port number must be an integer.");
					}
				}		
				
	    	Host [] returnHostArray = new Host[1];
	    	returnHostArray[0] = new Host( hostName, port );
			
			return returnHostArray;
			}
		catch( IOException e ) {
			System.out.println( "Exception occurred while reading user keyboard input: " + e.toString() );
			}
		return null;
		}
	
	
	
	/**
	 * Queries a list of suspected servers until a live one is found.<BR>
	 * Live server should send current list of live servers.
	 *
	 * @param suspectedOthers array of <code>Host</code>s that are believed to be alive.<BR>
	 * 
	 * @return array of live <code>Host</code>s that are returned by a live server, or 
	 * <code>null</code> if no live server connection is established
	 */
	protected Host[] lookForLiveServer( Host[] suspectedOthers ) {
		
		boolean found = false;
		NewServerUp m = new NewServerUp( localAddress );
		int i=0;
		
		ServerList sList = null;
		
		while( !found && i< suspectedOthers.length ) {
			
			Host h = suspectedOthers[i];
			if( ! h.equals( localAddress ) ) {	// make sure we don't try ourselves
				try {
					System.out.println( "Trying to connect to " + h.toString() );
					TwoWayMessageSocket sock = new TwoWayMessageSocket( h );
					
					System.out.println( "Asking server " + h.toString() + " for live server list." );
					sock.send( m );
					
					Message received = sock.receive();
					
					if( received instanceof ServerList ) {
						sList = (ServerList) received;
						found = true;
						System.out.println( "Live server list received." );
						}
					else {
						System.out.println("Bad message type (expecting ServerList) received from host "+ h.toString() );
						}	
					
					sock.close();
					}
				catch( IOException e ) {
					System.out.println("Connection to host failed: " + h.toString() );
					}
				}
			i++;
			}
		
		if( found ) {
			Vector hostVector = sList.getHostVector();
			
			localAddress = sList.getReceiverCorrectAddress();
			
			int size = hostVector.size();
			if( size != 0 ) {
				Host [] returnArray = new Host[ size ];
				for( int j=0; j<size; j++ ) {
					returnArray[j] = (Host)( hostVector.elementAt(j) );
					}
				return returnArray;
				}
			
			return null;
			}
		return null;
		}	// end lookForLiveServer()	
		
	
	
	
	
	
	
	
	/**
	 * Prints usage help message and exits.
	 */
	private static void usage() {				// print usage help message and exit

		System.out.print("usage:\n");
		System.out.print("Server port_number max_connect [new_chain]\n");
		
		System.out.print("examples:\n\t");
		System.out.print("Server 5433 15\n\t");
		System.out.print("Server 5433 15 0\n\t");
		System.out.print("Server 5433 15 1\n\t");
		
		System.exit(1);
		}

	
	
	/**
	 * <code>Thread</code> for receiving and processing initial messages from a 
	 * newly received socket connection.
 	 * <p>
 	 * This thread spawns other threads as appropriate.
 	 */
	class ReceiveThread extends Thread {

		Socket sock;

		TwoWayMessageSocket messageSock;
		
		/**
		 * Constructs a <code>ReceiveThread</code>/
		 *
		 * @param s <code>Socket</code> that this thread should receive messages on
		 */
		public ReceiveThread( Socket s ) {
			sock = s;
			}
		
		
		// overrides Thread run() function
		
		public void run() {
			
			try {
				messageSock = new TwoWayMessageSocket( sock );
				}
			catch( IOException e ) {
				System.out.println("Exception: " + e.toString() );
				System.out.println("Connection broken while receive thread setting up messsage socket.");
				return;
				}
			
			try {
				Message m = messageSock.receive();
				
				
				if( m instanceof ClientRequestConnect ) {
					System.out.println( "Client connection request message received." );
					
					// let ClientHandler decide what to do
					ClientHandlerThread cThread = new ClientHandlerThread( messageSock, (ClientRequestConnect)m );
					cThread.start();
					return;	// return without closing socket
					}
				else {
					// If we're the first server receiving the NewServerUp, i.e., directly
					// from the new server...
					if( m instanceof NewServerUp ) {
						// need to make sure correct address is in this message.
						// sometimes (if behind a firewall, e.g.), a server
						// won't retrieve it's local address correctly
						
						//NOTE:  Only fix address, not port, since socket port won't
						// be same port that server is listening on
						
						NewServerUp newUp = (NewServerUp)m;
						
						Host hostInMessage = newUp.getServerAddress();
						Host hostInSocket = messageSock.getRemoteHost();

						if( !hostInMessage.getAddress().equals( hostInSocket.getAddress() ) ) {
							hostInMessage.setAddress( hostInSocket.getAddress() );
							// need to fix host in message
							m = new NewServerUp( hostInMessage );
							}
							
						

						// now send new server the server list
						// include the possibly corrected local address for the server
						ServerList sList = new ServerList( propagator.getHostVector(), hostInMessage );
						messageSock.send( sList );
						// don't propagate the new server up message here...
						// that is handled below	
						}	
					
					boolean needsPropagated = true;
					boolean isPropMessage = false;
					if( m instanceof PropMessage ) {
						System.out.println( "Propagated message received." );
						
						isPropMessage = true;
						
						PropMessage pMessage = (PropMessage)m;
						// propagate to other servers
						propagator.propagateMessage( pMessage );
						
						// extract wrapped message and continue acting on it
						m = pMessage.getMessage();
						needsPropagated = false;
						}
					
					if( needsPropagated ) {
						propagator.propagateMessage( m );
						}
					
					if( m instanceof NewServerUp ) {
						
						NewServerUp newUp = (NewServerUp) m;
						
						System.out.println( "New server up message received:  " + newUp.getServerAddress() );
						
						// now add server to propagation list
						propagator.addServer( newUp.getServerAddress() );
						}
					
					else if( m instanceof ServerGoingDown ) {
						System.out.println( "Server going down message received." );
						
						ServerGoingDown goDown = (ServerGoingDown) m;
						propagator.removeServer( goDown.getServerAddress() );
						}
						
					else if( m instanceof FileHost ) {
						System.out.println( "File host file list message received." );
						
						FileHost fHost = (FileHost) m;
						database.addFileHost( fHost );
						}
						
					else if( m instanceof ClientDisconnect ) {
						ClientDisconnect cDis = (ClientDisconnect) m;
						FileHost fHost = new FileHost( cDis.getClientAddress(), new Vector() );
						database.removeFileHost( fHost );
						}
					}
				System.out.println( "Closing message socket." );
				messageSock.closeTotal();
				}
			catch( IOException e ) {
				System.out.println("Message receive failed in receive thread: " + e.toString() );
				e.printStackTrace();
				}
			}	// end run()

		}	// end ReceiveThread inner class	
	
	
	
	/**
	 * <code>Thread</code> for dealing with a client connection.
	 */
	class ClientHandlerThread extends Thread {
	
		TwoWayMessageSocket messageSock;
		
		Host connectedClient;
		
		ClientRequestConnect requestMessage;
		
		/**
		 * Constructs a <code>ClientHandlerThread</code>.
		 *
		 * @param mSock socket with connection to <code>Client</code>
		 * @param inRequestMessage message received from client requesting a connection
		 */
		public ClientHandlerThread( TwoWayMessageSocket mSock, ClientRequestConnect inRequestMessage ) {
			messageSock = mSock;
			requestMessage = inRequestMessage;
			}
		
		
		// overrides Thread run() function
		
		public void run() {
			
			try {
				
				Host hostInMessage = requestMessage.getClientAddress();
						
				Host hostInSocket = messageSock.getRemoteHost();
				
				if( !hostInMessage.getAddress().equals( hostInSocket.getAddress() ) ) {
					// client didn't send its local address correctly
							
					//NOTE:  Only fix address, not port, since socket port won't
					// be same port that server is listening on
							
					hostInMessage.setAddress( hostInSocket.getAddress() );
					}
				
				// first, send server list
				ServerList sList = new ServerList( propagator.getHostVector(), hostInMessage );
				sList.setStats( database.getNumClients(), database.getNumFiles(), database.getNumBytes() );	// add stats to message
				
				messageSock.send( sList );
				}
			catch( IOException e ) {
				System.out.println("Client connection failed.");
				return;	
				}
			
			// send accept or reject
			if( numClientsConnected >= maxClients ) {
				try {
					// reject client
					ConnectResponse response = new ConnectResponse( false );
					
					messageSock.send( response );
					
					messageSock.closeTotal();
					}
				catch( IOException e ) {
					System.out.println("Client connection failed.");
					return;	
					}
				}
			
			else {
				// there's room for client
				numClientsConnected++;
				
				// so we can remove client from DB on disconnect
				FileHost savedFileHost;
				
				ConnectResponse response = new ConnectResponse( true );
				
				try {
					messageSock.send( response );
					
					
					// now receive client file list
					
					Message m = messageSock.receive();
					
					if( m instanceof FileHost ) {
						FileHost fHost = (FileHost) m;
						
						connectedClient = fHost.getHost();
						
						Host hostInSocket = messageSock.getRemoteHost();
						
						/*
							This problem has been solved above by sending client
							its correct address along with the server list.
							
						if( !connectedClient.getAddress().equals( hostInSocket.getAddress() ) ) {
							// client didn't send its local address correctly
							
							//NOTE:  Only fix address, not port, since socket port won't
							// be same port that server is listening on
							
							connectedClient.setAddress( hostInSocket.getAddress() );
							
							// need to run through vector and replace all file descriptors
							// so that they all have correct client address
							Vector oldFileVector = fHost.getFileVector();

							Vector newFileVector = new Vector( oldFileVector.size() );
							
							int numFiles = oldFileVector.size();
							
							for( int i=0; i<numFiles; i++ ) {
								konspire.common.FileDescriptor oldFD = 
											(konspire.common.FileDescriptor)( oldFileVector.elementAt(i) );	
								konspire.common.FileDescriptor newFD = 
											new konspire.common.FileDescriptor( connectedClient,
												oldFD.getName(), oldFD.getPath(), oldFD.getSize() );
								newFileVector.addElement( newFD );
								}
							
							}
						*/
						
							
						// propagate first, then add to DB
						// since DB add changes fhost object
						propagator.propagateMessage( fHost );
						database.addFileHost( fHost );
						savedFileHost = fHost;
						}
					else {
						System.out.println("Client sent an out of order message.");
						messageSock.closeTotal();
						return;
						}
					}
				catch( IOException e ) {
					System.out.println("Client connection failed.");
					numClientsConnected--;
					return;	
					}
				
				
				// now go into main client loop
			
				boolean connected = true;
				try {
					while( connected ) {
						Message m = messageSock.receive();
						
						if( m instanceof FileHost ) {
							// a new copy of client file list arrived
							FileHost fHost = (FileHost) m;
							// propagate first, then add to DB
							// since DB add changes fhost object
							propagator.propagateMessage( fHost );
							database.addFileHost( fHost );
							}
						else if( m instanceof ClientDisconnect ) {
							ClientDisconnect cDis = (ClientDisconnect) m;
							database.removeFileHost( savedFileHost );
							propagator.propagateMessage( cDis );
							messageSock.closeTotal();
							
							connected = false;
							}
						else if( m instanceof SearchQuery ) {
							SearchQuery sQ = (SearchQuery) m;
							Vector results = database.searchForStrings( sQ.getSearchTerms(), sQ.getNumResults(), connectedClient );
							SearchResults sR = new SearchResults( results, sQ );
							sR.setStats( database.getNumClients(), database.getNumFiles(), database.getNumBytes() );	// add stats to message
							messageSock.send( sR );
							}
						else {
							System.out.println("Unknown message type received from client.");
							}
						}
					}
				catch( IOException e ) {
					System.out.println("Client connection failed.");
					database.removeFileHost( savedFileHost );
					Host clientAddress = savedFileHost.getHost();
					propagator.propagateMessage( new ClientDisconnect( clientAddress ) );
					}
				
				// client has disconnected if we get to this point
				numClientsConnected--;
				}	
			}	// end run()

		}	// end ClientHandlerThread inner class
	
	
		
	}	// end Server class
