// Jason Rohrer
// ClientDatabase.java

/**
*
*	Database of client hosts with files
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	6-14-2000	Added support for passing client host name into 
*									StringTree for file search. (used to filter results)
*/

package konspire.server.database;

import konspire.common.FileHost;
import konspire.common.FileDescriptor;

import konspire.common.Host;

import konspire.common.Vector;
import java.util.Enumeration;


/**
 * Database of client file hosts.
 *
 * @author Jason Rohrer
 */
public class ClientDatabase {
	
	// vector of fileHosts
	Vector fileHostVector;
	
	StringTree tree;
	
	
	long filesOnline = 0;
	long bytesOnline = 0;
	long clientsOnline = 0;
	
	
	/**
	 * Constructs a <code>ClientDatabase</code>.
	 */
	public ClientDatabase() {
		tree = new StringTree();
		
		fileHostVector = new Vector();
		}
	
	
	/**
	 * Adds a file host to the database.
	 *
	 * @param fHost the file host to add
	 */
	public synchronized void addFileHost( FileHost fHost ) {

		if( fileHostVector.contains( fHost ) ) {
			// remove old copy of FileHost if it exists
			removeFileHost( fHost );
			}

		fileHostVector.addElement( fHost );


		// vector that will eventually repace file vector in fHost
		// will contain StringTreeFileDescriptors instead of FileDescriptors
		Vector newFileVector = new Vector();
			
		// run through fileDescriptors, inserting them into tree
		Enumeration en = fHost.getFileVector().elements();
		while( en.hasMoreElements() ) {
			FileDescriptor nextFile = (FileDescriptor)( en.nextElement() );
			
			// insert into tree, and get StringTreeFileDescriptor for file
			StringTreeFileDescriptor treeFileWrapper = tree.insertFile( nextFile );
			
			newFileVector.addElement( treeFileWrapper );
			
			filesOnline ++;
			bytesOnline += nextFile.getSize();
			}
		
		
		fHost.setFileVector( newFileVector );
		
		// now can look up files in tree with direct node pointers
		// (good for file removal)
		
		clientsOnline++;
		}
	
	
	/**
	 * Removes a file host from the database.
	 *
	 * @param fHost the file host to remove
	 */	
	public synchronized void removeFileHost( FileHost fHost ) {
		
		// first, need to find a host that matches this host in OUR host vector
		boolean found = false;
		Enumeration e = fileHostVector.elements();
		while( !found && e.hasMoreElements() ) {
			FileHost other = (FileHost)( e.nextElement() );
			if( other.equals( fHost ) ) {
				found = true;
				fHost = other;
				}
			}
		
		if( !found ) {
			return;
			}
		
		
		// run through StringTreeFileDescriptors, removing them from tree
		e = fHost.getFileVector().elements();
		while( e.hasMoreElements() ) {
			StringTreeFileDescriptor nextFile = (StringTreeFileDescriptor)( e.nextElement() );
			
			StringTreeNode node = nextFile.getNode();
			
			// Convert to a FileDescriptor and remove from tree
			node.removeFile( nextFile.toFileDescriptor() );
			filesOnline --;
			bytesOnline -= nextFile.getSize();
			}
		
		fileHostVector.removeElement( fHost );		// remove from client DB host vector
		clientsOnline--;
		}
	
	
	/**
	 * Search database for files matching a collection of search terms.
	 * Results are filtered so that no files on searching client are returned.
	 * 
	 * @param searchTerms an array of terms to match in search
	 * @param numResults maximum number of results to return
	 * @param searchingClient the client performing this search
	 *
	 * @return a <code>Vector</code> of results
	 */
	public synchronized Vector searchForStrings( String[] searchTerms, int numResults, Host searchingClient ) {
		return tree.searchForStrings( searchTerms, numResults, searchingClient );
		}
	
	
	/**
	 * Gets number of files indexed by the database.
	 *
	 * @return number of files indexed by the databse
	 */
	public long getNumFiles() {
		return filesOnline;
		}
	
	
	/**
	 * Gets number of bytes in files indexed by the databse.
	 *
	 * @return number of bytes in files indexed by the databse
	 */
	public long getNumBytes() {
		return bytesOnline;
		}
	
	
	/**
	 * Gets number <code>Client</code>s that have files indexed by the database.
	 *
	 * @return number <code>Client</code>s that have files indexed by the database
	 */
	public long getNumClients() {
		return clientsOnline;
		}
		
	}

